"""
This is the RiCA Connector whose purpose is to connect the RiCA Server with the LLM.
"""

# Base adapter class
from .base import ReasoningThreadBase

# ---- Adapters ----

# Transformers Adapter (requires `pip install .[pt]`)
try:
    from . import transformers_adapter as transformer_adapter  # alias for compatibility
    from . import transformers_adapter as transformers_adapter
except ImportError:

    class _MissingAdapter:
        def __getattr__(self, name):
            raise ImportError(
                "The transformers adapter requires 'transformers' and 'torch'. "
                "Please install them with `pip install .[pt]`."
            )

    transformers_adapter = _MissingAdapter()
    transformer_adapter = _MissingAdapter()

# Future adapters (e.g., OpenAI, Anthropic) can be added here in a similar pattern.

__all__ = ["ReasoningThreadBase", "transformers_adapter", "transformer_adapter"]



import json
import logging
import re
from typing import Any, Tuple
from xml.etree import ElementTree as ET

from rica.exceptions import InvalidRiCAString

logger = logging.getLogger(__name__)


def parse_rica_tag(tag_text: str) -> Tuple[str, str, Any]:
    """
    Parses a single <rica> tag and returns (package, route, content).

    Args:
        tag_text: The raw string of the <rica> tag.

    Returns:
        A tuple containing (package_name, route_name, content).
        Content is a dict or list if JSON, or a string.

    Raises:
        InvalidRiCAString: If the tag is malformed or missing attributes.
    """
    try:
        # Attempt to parse XML
        try:
            root = ET.fromstring(tag_text)
        except ET.ParseError as e:
            logger.error(f"XML parse error: {e}, tag_text: {tag_text[:200]}")
            # Fallback regex
            package_match = re.search(r'package=["\']([^"\']+)["\']', tag_text)
            route_match = re.search(r'route=["\']([^"\']+)["\']', tag_text)

            if package_match and route_match:
                package_name = package_match.group(1)
                route_name = route_match.group(1)
                content_match = re.search(r">\s*(.*?)\s*<\/rica>", tag_text, re.DOTALL)
                content_str = content_match.group(1) if content_match else ""

                class MockRoot:
                    attrib = {"package": package_name, "route": route_name}
                    text = content_str

                root = MockRoot()
            else:
                raise InvalidRiCAString(f"Invalid XML format: {e}")

        package_name = root.attrib.get("package")
        route_name = root.attrib.get("route")

        if not package_name or not route_name:
            raise InvalidRiCAString("Missing package or route attribute")

        content_str = root.text or ""
        content = json.loads(content_str) if content_str.strip() else {}

        return package_name, route_name, content

    except Exception as e:
        if isinstance(e, InvalidRiCAString):
            raise
        raise InvalidRiCAString(f"Error parsing RiCA tag: {e}") from e



import importlib.util
import os
import sys
import tarfile
import tempfile

from rica.core.application import RiCA


async def load_app_from_path(path: str) -> RiCA:
    """
    Helper to load RiCA app from a file path.

    Args:
        path: Path to a .py file or .tar.gz/.tar archive.

    Returns:
        The loaded RiCA application instance.

    Raises:
        ImportError: If loading fails or no RiCA instance is found.
    """
    if path.endswith(".tar.gz") or path.endswith(".tar"):
        # Handle archive
        temp_dir = tempfile.mkdtemp()
        try:
            with tarfile.open(path, "r:*") as tar:
                tar.extractall(path=temp_dir)

            # Look for a python package or module in the extracted content
            # We assume there's an __init__.py or a .py file
            # Simple heuristic: add temp_dir to sys.path and try to import
            sys.path.insert(0, temp_dir)

            # Find what to import. We look for the first directory with __init__.py
            # or the first .py file. This is a bit ambiguous. Let's assume the
            # archive contains a folder with the package name OR the archive content
            # IS the package.

            # Let's try to find a .py file or a directory
            found_module = None
            for root, dirs, files in os.walk(temp_dir):
                for file in files:
                    if file.endswith(".py"):
                        # If it's __init__.py, the package is the parent dir name
                        if file == "__init__.py":
                            found_module = os.path.basename(root)
                            break
                        else:
                            # It's a module file
                            found_module = file[:-3]
                            break
                if found_module:
                    break

            if not found_module:
                raise ImportError("Could not find a valid Python module or package in the archive.")

            try:
                module = importlib.import_module(found_module)
            finally:
                # Clean up sys.path? Maybe not if we want to keep it loaded.
                # But for this session it might be fine.
                # Ideally we should remove it to avoid pollution, but we need the module code.
                pass

        except Exception as e:
            raise ImportError(f"Failed to load app from archive '{path}': {e}")

    elif path.endswith(".py"):
        # Handle single .py file
        module_name = os.path.basename(path)[:-3]
        spec = importlib.util.spec_from_file_location(module_name, path)
        if not spec or not spec.loader:
            raise ImportError(f"Could not load module from '{path}'")
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
    else:
        raise ImportError(f"Unsupported file type: {path}")

    # Find RiCA instance in module
    # 1. Look for 'app' variable
    if hasattr(module, "app") and isinstance(module.app, RiCA):
        return module.app

    # 2. Look for any RiCA instance
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if isinstance(attr, RiCA):
            return attr

    raise ImportError(f"No 'RiCA' instance found in module loaded from '{path}'.")



from typing import Dict

from jinja2 import Template

from ..core.application import RiCA

# Using Jinja2 template for more flexibility
SYSTEM_PROMPT_TEMPLATE = Template(
    """
You are RiCA {{ model_name }}, powered by {{ model_modal }}.

Unlike traditional assistants, you can call tools and continue thinking while waiting for callbacks.

**IMPORTANT**: Only content within <rica> blocks will be processed by RiCA Server.
All other text is your internal reasoning and will NOT be shown to the user.

## Available Tools

{% for package_name, app in apps.items() %}
### Application: {{ package_name }}
Description: {{ app.description or "No description" }}

{% for route in app.routes %}
- **Route**: `{{ route.route }}`
  - **Description**: {{ route.function.__doc__ or "No description" }}
  - **Background**: {{ route.background }}
  {% if route.background %}
  - **Timeout**: {{ route.timeout }}ms
  {% endif %}
{% endfor %}
{% endfor %}

## Usage Guide

To call a tool, use this format:

```xml
<rica package="package_name" route="/route_name">
{
  "arg1": "value1",
  "arg2": "value2"
}
</rica>
```

Example:

```xml
<rica package="demo.sys" route="/exec">
{
  "code": "1 + 1"
}
</rica>
```

To send a response to the user:

```xml
<rica package="rica" route="/response">
[
  {
    "type": "text",
    "content": "This is the final answer to the user."
  }
]
</rica>
```
"""
)


async def _rica_prompt(apps: Dict[str, RiCA], model_name: str, model_modal: str) -> str:
    """Generate system prompt using Jinja2 template."""
    return SYSTEM_PROMPT_TEMPLATE.render(apps=apps, model_name=model_name, model_modal=model_modal)



"""Custom exceptions for RiCA Server."""


class PackageInvalidError(ValueError):
    """Raised when a package name is syntactically invalid."""


class PackageExistError(ValueError):
    """Raised when attempting to register a package that already exists."""


class RouteExistError(ValueError):
    """Raised when attempting to register a route that already exists in a package."""


class RouteNotFoundError(ValueError):
    """Raised when a requested route cannot be found in the package."""


class PackageNotFoundError(ValueError):
    """Raised when a requested package cannot be found in the registry."""


class ExecutionTimedOut(Exception):
    """Raised when a routed tool call times out."""


class UnexpectedExecutionError(Exception):
    """Raised when a routed tool call fails unexpectedly."""


class InvalidRiCAString(ValueError):
    """Raised when an input tool-call string is malformed or invalid."""


class AdapterDependenciesImportError(ImportError):
    """Raised when optional adapter dependencies are missing."""



import logging
from typing import Dict

logger = logging.getLogger(__name__)


class Whiteboard:
    """
    A shared whiteboard for storing context and data across threads.
    """

    def __init__(self):
        self._content: Dict[str, str] = {}
        self._descriptions: Dict[str, str] = {}

    def handle(self, input_data: dict) -> dict:
        """
        Handle whiteboard operations.

        Input Schema:
        {
            "action": "read" | "write" | "append" | "clear",
            "whiteboard_id": "string",
            "content": "string (optional)",
            "description": "string (optional)"
        }
        """
        action = input_data.get("action")
        wb_id = input_data.get("whiteboard_id", "default")
        content = input_data.get("content", "")
        description = input_data.get("description")

        if not action:
            return {"error": "Missing 'action' parameter"}

        if action == "read":
            return {
                "content": self._content.get(wb_id, ""),
                "description": self._descriptions.get(wb_id, ""),
            }

        elif action == "write":
            self._content[wb_id] = content
            if description is not None:
                self._descriptions[wb_id] = description
            return {"status": "success", "message": f"Whiteboard '{wb_id}' updated"}

        elif action == "append":
            current = self._content.get(wb_id, "")
            self._content[wb_id] = current + "\n" + content
            if description is not None:
                self._descriptions[wb_id] = description
            return {"status": "success", "message": f"Appended to whiteboard '{wb_id}'"}

        elif action == "clear":
            self._content.pop(wb_id, None)
            self._descriptions.pop(wb_id, None)
            return {"status": "success", "message": f"Whiteboard '{wb_id}' cleared"}

        elif action == "list":
            return {
                "whiteboards": [
                    {"id": wid, "description": self._descriptions.get(wid, "")}
                    for wid in self._content.keys()
                ]
            }

        else:
            return {"error": f"Unknown action: {action}"}


# Singleton instance
_whiteboard_instance = Whiteboard()


def get_whiteboard():
    return _whiteboard_instance



from __future__ import annotations

import asyncio
import json
from typing import Any, Dict, Optional

from ..exceptions import AdapterDependenciesImportError
from ..utils.prompt import _rica_prompt
from .base import ReasoningThreadBase

try:
    import torch
    from transformers import (
        AutoModelForCausalLM,
        AutoTokenizer,
        GenerationConfig,
        StoppingCriteria,
        StoppingCriteriaList,
        TextIteratorStreamer,
    )
except ImportError:
    raise AdapterDependenciesImportError(
        "The transformers adapter requires 'transformers' and 'torch' to be installed."
    )

from ..config import DEFAULT_MODEL_NAME

default_model_name = DEFAULT_MODEL_NAME


class _ToolCallStoppingCriteria(StoppingCriteria):
    """Stops generation when a complete <rica>...</rica> tag is detected."""

    def __init__(self, tokenizer: AutoTokenizer):
        self.tokenizer = tokenizer

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs) -> bool:
        full_text = self.tokenizer.decode(input_ids[0], skip_special_tokens=True)
        if full_text.rstrip().endswith("</rica>"):
            return True
        return False


class ReasoningThread(ReasoningThreadBase):
    """
    A reasoning thread based on Hugging Face Transformers that supports a token-by-token
    generation loop with real-time text insertion and tool-call execution.
    """

    def __init__(
        self,
        context: str = "",
        model_name: str = default_model_name,
        generation_config: Optional[Dict[str, Any]] = None,
        model: Optional[AutoModelForCausalLM] = None,
        tokenizer: Optional[AutoTokenizer] = None,
    ):
        super().__init__(context)
        self.model_name: str = model_name
        self.model_modal: str = "PyTorch/Transformers"

        self._task: Optional[asyncio.Task] = None
        self._lock = asyncio.Lock()
        self._prompt_injected = False

        self._pause_event = asyncio.Event()
        self._stop_event = asyncio.Event()
        self._is_running = False
        self._pause_event.set()

        self._model: Optional[AutoModelForCausalLM] = model
        self._tokenizer: Optional[AutoTokenizer] = tokenizer
        self._streamer: Optional[TextIteratorStreamer] = None
        self._stopping_criteria: Optional[StoppingCriteriaList] = None

        default_config = {
            "max_new_tokens": 1024,
            "do_sample": True,
            "temperature": 0.6,
            "top_p": 0.9,
        }
        if generation_config:
            default_config.update(generation_config)
        self._generation_config = GenerationConfig(**default_config)

    async def create_sub_thread(
        self, model_name: Optional[str] = None, config: Optional[Dict[str, Any]] = None
    ) -> Any:
        """
        Create a sub-thread.
        If model_name is None or matches current model, share the model instance.
        """
        target_model_name = model_name or self.model_name

        shared_model = None
        shared_tokenizer = None

        if target_model_name == self.model_name:
            await self._ensure_model()
            shared_model = self._model
            shared_tokenizer = self._tokenizer

        return ReasoningThread(
            model_name=target_model_name,
            generation_config=config,
            model=shared_model,
            tokenizer=shared_tokenizer,
        )

    # --------------------------------------------------------------------------
    # Public Lifecycle API
    # --------------------------------------------------------------------------

    async def insert(self, text: Any):
        if text is None:
            return

        # Ensure model is loaded and prompt is injected before processing input
        await self._ensure_model()
        await self._inject_prompt_if_needed()

        s = text if isinstance(text, str) else json.dumps(text, ensure_ascii=False)
        formatted_input = f'<rica-callback package="rica.userinput">{s}</rica-callback>'

        self.pause()
        async with self._lock:
            self._context += formatted_input
        await self._emit_token(formatted_input)

        self.run()

    async def wait(self):
        if self._task and not self._task.done():
            await self._task

    async def destroy(self):
        if self._task and not self._task.done():
            self._stop_event.set()
            self._pause_event.set()
            try:
                await asyncio.wait_for(self._task, timeout=5.0)
            except (asyncio.CancelledError, asyncio.TimeoutError):
                self._task.cancel()

    def run(self):
        self._pause_event.set()
        if not self._is_running:
            self._is_running = True
            self._stop_event.clear()
            self._task = asyncio.create_task(self._run_loop())

    def pause(self):
        self._pause_event.clear()

    # --------------------------------------------------------------------------
    # Internal Helper Methods
    # --------------------------------------------------------------------------

    def _get_next_token_from_streamer(self) -> Optional[str]:
        """
        Synchronous helper to safely get the next token from the streamer.
        It catches StopIteration and returns None, preventing the exception
        from propagating into the asyncio event loop.
        """
        try:
            return next(self._streamer)
        except StopIteration:
            return None

    async def _ensure_model(self):
        if self._model and self._tokenizer:
            return

        def _load_sync():
            tokenizer = AutoTokenizer.from_pretrained(self.model_name)

            dtype = torch.bfloat16 if torch.cuda.is_available() else torch.float32

            # 使用 device_map="auto" 直接将模型加载到 GPU，避免 CPU 内存溢出
            model = AutoModelForCausalLM.from_pretrained(
                self.model_name,
                low_cpu_mem_usage=True,
                device_map="auto",
                torch_dtype=dtype,
            )

            if tokenizer.pad_token is None:
                tokenizer.pad_token = tokenizer.eos_token

            return model, tokenizer

        self._model, self._tokenizer = await asyncio.to_thread(_load_sync)
        self._streamer = TextIteratorStreamer(
            self._tokenizer,
            skip_prompt=True,
            skip_special_tokens=True,
            timeout=None,
        )
        self._stopping_criteria = StoppingCriteriaList([_ToolCallStoppingCriteria(self._tokenizer)])
        self._model.eval()

    async def _inject_prompt_if_needed(self):
        if self._prompt_injected:
            return

        async with self._lock:
            async with self._apps_lock:
                system_prompt = await _rica_prompt(self._apps, self.model_name, self.model_modal)
            self._context = system_prompt + self._context
            self._prompt_injected = True

    async def _run_loop(self):
        try:
            await self._ensure_model()

            while not self._stop_event.is_set():
                await self._pause_event.wait()
                if self._stop_event.is_set():
                    break

                await self._inject_prompt_if_needed()

                async with self._lock:
                    inputs = self._tokenizer(self._context, return_tensors="pt").to(
                        self._model.device
                    )

                generation_kwargs = {
                    "input_ids": inputs["input_ids"],
                    "attention_mask": inputs["attention_mask"],
                    "generation_config": self._generation_config,
                    "streamer": self._streamer,
                    "stopping_criteria": self._stopping_criteria,
                }

                async def _read_from_streamer():
                    """Read tokens from streamer and emit them."""
                    while True:
                        try:
                            new_text = await asyncio.to_thread(self._get_next_token_from_streamer)

                            if new_text is None:
                                break

                            if self._stop_event.is_set() or not self._pause_event.is_set():
                                break

                            async with self._lock:
                                self._context += new_text
                            await self._emit_token(new_text)

                        except StopIteration:
                            break

                generation_task = asyncio.to_thread(self._model.generate, **generation_kwargs)
                reader_task = asyncio.create_task(_read_from_streamer())

                await asyncio.gather(generation_task, reader_task, return_exceptions=True)

                was_executed, _ = await self._detect_and_execute_tool_tail()
                if not was_executed:
                    self.pause()

        except Exception as e:
            error_msg = f"[adapter-error]{type(e).__name__}: {e}"
            await self._emit_token(error_msg)
            if not isinstance(e, asyncio.CancelledError):
                raise
        finally:
            self._is_running = False



import logging
import uuid
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)


class ThreadManager:
    """
    Manages the lifecycle of sub-threads.
    """

    def __init__(self):
        self._threads: Dict[str, Any] = {}  # id -> ReasoningThread instance

    def register_thread(self, thread_id: str, thread_instance: Any):
        self._threads[thread_id] = thread_instance

    def unregister_thread(self, thread_id: str):
        self._threads.pop(thread_id, None)

    def get_thread(self, thread_id: str) -> Optional[Any]:
        return self._threads.get(thread_id)

    def list_threads(self) -> Dict[str, str]:
        return {tid: str(t) for tid, t in self._threads.items()}

    async def spawn(self, parent_thread: Any, input_data: dict) -> dict:
        """
        Spawn a new sub-thread.

        Input:
        {
            "task": "string - initial prompt",
            "model": "string - optional model name",
            "config": "dict - optional generation config"
        }
        """
        task = input_data.get("task")
        if not task:
            return {"error": "Missing 'task' parameter"}

        model_name = input_data.get("model")
        config = input_data.get("config")

        # Create a new thread instance
        # We need to import ReasoningThread dynamically to avoid circular imports
        # or rely on a factory. For now, we assume the parent thread has a factory method.
        if not hasattr(parent_thread, "create_sub_thread"):
            return {"error": "Parent thread does not support spawning sub-threads"}

        try:
            sub_thread = await parent_thread.create_sub_thread(model_name=model_name, config=config)
            thread_id = str(uuid.uuid4())
            self.register_thread(thread_id, sub_thread)

            # Start the thread
            await sub_thread.initialize()
            sub_thread.run()

            # Insert the task
            await sub_thread.insert(task)

            return {"status": "success", "thread_id": thread_id, "message": "Thread spawned"}
        except Exception as e:
            logger.error(f"Failed to spawn thread: {e}", exc_info=True)
            return {"error": f"Failed to spawn thread: {str(e)}"}

    async def kill(self, input_data: dict) -> dict:
        thread_id = input_data.get("thread_id")
        if not thread_id:
            return {"error": "Missing 'thread_id'"}

        thread = self.get_thread(thread_id)
        if not thread:
            return {"error": f"Thread '{thread_id}' not found"}

        await thread.destroy()
        self.unregister_thread(thread_id)
        return {"status": "success", "message": f"Thread '{thread_id}' killed"}

    def list(self, input_data: dict) -> dict:
        return {"threads": self.list_threads()}


# Singleton
_thread_manager = ThreadManager()


def get_thread_manager():
    return _thread_manager



import asyncio
import functools
import json
import logging
import re
from typing import Any, Callable, Coroutine, Dict, List, Optional, Union
from uuid import uuid4

from rica.core.application import CallBack, RiCA, Route
from rica.exceptions import (
    ExecutionTimedOut,
    PackageExistError,
    PackageNotFoundError,
    RouteNotFoundError,
    UnexpectedExecutionError,
)
from rica.utils.package_loader import load_app_from_path
from rica.utils.parser import parse_rica_tag

__all__ = ["ReasoningThreadBase"]

# Configure logger
logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)


class ReasoningThreadBase:
    """
    Base template for reasoning adapters.

    This class provides reusable, framework-agnostic helpers for:
    - Managing a textual context buffer.
    - Managing a collection of RiCA applications.
    - Registering two types of callbacks:
      - @token_generated: Receives every raw token from the model.
      - @trigger: Receives only the payload from a rica.response tool call.
    - Detecting and executing <rica ...>...</rica> tool calls.

    Attributes:
        _apps: A dictionary of installed RiCA applications.
        _context: The textual context buffer.
    """

    def __init__(self, context: str = ""):
        """
        Initializes the reasoning thread.

        Note:
            The instance must be initialized by calling `await instance.initialize()`
            before any other operations.

        Args:
            context: The initial context string.
        """
        self._apps: Dict[str, RiCA] = {}
        self._apps_lock = asyncio.Lock()
        self._context: str = context or ""
        self._response_callbacks: List[Callable[[Any], Any]] = []  # For @trigger
        self._token_callbacks: List[Callable[[str], Any]] = []  # For @token_generated
        self._initialized = False

    async def initialize(self):
        """Initialize the reasoning thread by installing system apps."""
        if self._initialized:
            return
        # Install the virtual 'rica' app for system prompts
        await self.install(RiCA("rica"))
        self._initialized = True

    async def install(self, app: Union[RiCA, str]):
        """
        Installs a RiCA application.

        Args:
            app: An instance of the RiCA class, or a file path (str) to a .py file
                 or .tar.gz archive.

        Raises:
            TypeError: If the 'app' argument is invalid.
            PackageExistError: If an application with the same package name is
                               already installed.
            ImportError: If loading from file fails.
        """
        if isinstance(app, str):
            # Dynamic loading logic
            app_instance = await load_app_from_path(app)
        elif isinstance(app, RiCA):
            app_instance = app
        else:
            raise TypeError("The 'app' argument must be an instance of RiCA or a file path string.")

        async with self._apps_lock:
            if app_instance.package in self._apps:
                raise PackageExistError(
                    f"Application with package '{app_instance.package}' is already installed."
                )
            self._apps[app_instance.package] = app_instance

    async def uninstall(self, package_name: str):
        """
        Uninstalls a RiCA application by its package name.

        Args:
            package_name: The package name of the application to uninstall.

        Raises:
            PackageNotFoundError: If the application with the given package name is not found.
        """
        async with self._apps_lock:
            if package_name not in self._apps:
                raise PackageNotFoundError(f"Application with package '{package_name}' not found.")
            del self._apps[package_name]

    # ---- Lifecycle placeholders (to be implemented by subclasses) ----
    async def insert(self, text: Any):
        """Insert external text into the context (to be implemented by subclass)."""
        raise NotImplementedError

    async def wait(self):
        """Wait for current reasoning/generation to complete (to be implemented)."""
        raise NotImplementedError

    async def destroy(self):
        """Stop and clean up resources (to be implemented)."""
        raise NotImplementedError

    def run(self):
        """Start or resume reasoning (to be implemented)."""
        raise NotImplementedError

    def pause(self):
        """Pause reasoning (to be implemented)."""
        raise NotImplementedError

    @property
    def context(self) -> str:
        """Return the current textual context buffer."""
        return self._context

    # ---- Common helpers ----
    def trigger(self, function: Callable[..., Any]) -> Callable[..., Any]:
        """
        Register a callback that will be called for `rica.response` tool calls.
        This is for delivering final responses to the user.
        """
        self._response_callbacks.append(function)
        return function

    def token_generated(self, function: Callable[[str], Any]) -> Callable[[str], Any]:
        """
        Register a callback that will be called for every token generated by the model.
        This is for observing the model's "thinking" process.
        """
        self._token_callbacks.append(function)
        return function

    async def _emit_response(self, payload: Any):
        """Emit a final response payload to all @trigger callbacks."""
        if not payload:
            return
        tasks = [
            asyncio.create_task(cb(payload))
            for cb in self._response_callbacks
            if asyncio.iscoroutinefunction(cb)
        ]
        tasks.extend(
            [
                asyncio.create_task(asyncio.to_thread(cb, payload))
                for cb in self._response_callbacks
                if not asyncio.iscoroutinefunction(cb)
            ]
        )
        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

    async def _emit_token(self, piece: str):
        """Emit a raw generated token to all @token_generated callbacks."""
        if not piece:
            return
        tasks = [
            asyncio.create_task(cb(piece))
            for cb in self._token_callbacks
            if asyncio.iscoroutinefunction(cb)
        ]
        tasks.extend(
            [
                asyncio.create_task(asyncio.to_thread(cb, piece))
                for cb in self._token_callbacks
                if not asyncio.iscoroutinefunction(cb)
            ]
        )
        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

    async def _execute_tool_call(
        self, app: Route, data: list | dict
    ) -> Union[CallBack, Coroutine[Any, Any, None]]:
        """
        Execute a registered function.

        Args:
            app: The Route object representing the tool to be executed.
            data: The input data for the tool.

        Returns:
            A CallBack object for synchronous tools, or a coroutine for background tools.
        """
        function = app.function
        timeout = app.timeout
        background = app.background

        start_time = asyncio.get_event_loop().time()

        try:
            if not asyncio.iscoroutinefunction(function):
                function = asyncio.to_thread(functools.partial(function, data))
            else:
                function = function(data)

            if background:
                call_id = uuid4()
                loop = asyncio.get_running_loop()
                task = loop.create_task(function, name=str(call_id))

                def on_task_done(t):
                    try:
                        if t.cancelled():
                            logger.warning(f"Task {call_id} was cancelled")
                        elif t.exception():
                            logger.error(f"Task {call_id} failed", exc_info=t.exception())
                    except Exception as e:
                        logger.error(f"Error in task callback: {e}")

                task.add_done_callback(on_task_done)

                if timeout > 0:
                    cancel_handle = loop.call_later(timeout / 1000, task.cancel)
                    task.add_done_callback(lambda _: cancel_handle.cancel())
                return call_id
            else:
                try:
                    result = (
                        await asyncio.wait_for(function, timeout / 1000)
                        if timeout > 0
                        else await function
                    )
                    duration = (asyncio.get_event_loop().time() - start_time) * 1000
                    return CallBack(
                        package=app.route.split("/")[0],
                        route=app.route,
                        call_id=uuid4(),
                        callback=result,
                        duration_ms=duration,
                    )
                except asyncio.TimeoutError as e:
                    logger.error(f"Tool call timed out after {timeout}ms")
                    raise ExecutionTimedOut from e
                except Exception as e:
                    logger.error(f"Tool execution failed: {e}", exc_info=True)
                    raise UnexpectedExecutionError(str(e)) from e
        except Exception as e:
            logger.critical(f"Critical error in _execute_tool_call: {e}", exc_info=True)
            raise

    async def create_sub_thread(
        self, model_name: Optional[str] = None, config: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Create a sub-thread. To be implemented by subclasses."""
        raise NotImplementedError

    async def _detect_and_execute_tool_tail(self) -> tuple[bool, Optional[str]]:
        """Detect and execute all <rica ...>...</rica> tags in the context."""
        # Find all tags
        pattern = r"<rica\s+[^>]*>.*?<\/rica>"
        matches = list(re.finditer(pattern, self._context, re.DOTALL | re.IGNORECASE))

        if not matches:
            return False, None

        # Only process tags that haven't been processed yet.
        # Since we append results to the context, we need a way to distinguish processed tags.
        # A simple heuristic is to check if the tag is followed by a result block
        # or if it's at the very end.
        # However, simpler is to assume the model generates tags at the end.
        # But if we have multiple tags, we want to process them all.

        # Strategy: Find the last contiguous block of tags at the end of the context.
        # This is tricky because there might be whitespace.

        # Simplified Strategy:
        # 1. Find all tags.
        # 2. Filter out tags that already have a corresponding result (this is hard without IDs).
        # 3. Actually, the standard flow is:
        #    Model generates -> Stop at </rica> -> We process -> Append result.
        # If the model generates multiple tags in one go (e.g. <rica>A</rica> <rica>B</rica>),
        # the stopping criteria might stop at the first </rica> or the last?
        # The current stopping criteria stops at ANY </rica>.
        # So we likely only have ONE tag at the end.
        # BUT, if the model is fast or we change stopping criteria, we might have multiple.
        # Let's assume we process the LAST tag found.
        # Wait, user wants "simultaneous tool calls".
        # If the model outputs: <rica>A</rica>\n<rica>B</rica>
        # We should execute both.

        # Let's try to find all tags that are NOT followed by a result block (e.g. JSON or UUID).
        # This is getting complex.
        # Alternative: The `_context` grows. We can keep track of the `last_processed_index`.

        # For now, let's stick to the "tail" concept but expand it to "tail block".
        # We look for the last occurrence of a tag.
        # If we want parallel, we need to find ALL tags in the newly generated text.
        # But `_detect_and_execute_tool_tail` is called after generation stops.
        # If generation stops at `</rica>`, we might only have one tag.
        # UNLESS we change stopping criteria to NOT stop, or the model outputs multiple
        # tags quickly.

        # Let's assume the model outputs one tag at a time for now, as per the stopping criteria.
        # To support parallel, the model would need to output multiple tags BEFORE stopping.
        # If we want to support that, we need to adjust `_ToolCallStoppingCriteria` in
        # `transformers_adapter.py`.

        # For this step, let's just make `_detect_and_execute_tool_tail` robust enough
        # to handle multiple tags if they exist.

        # We will process ALL tags found at the end of the string.
        # We iterate backwards? No.

        # Let's just process the last match for now to maintain stability,
        # but if we want parallel, we should process all matches that don't look like
        # they have a result.
        # Given the complexity and the "Demo" nature, let's implement a robust
        # "process all new tags" logic.
        # We can assume "new tags" are those after the last known context length?
        # `ReasoningThread` doesn't track "last known length" explicitly in `base.py`.

        # Let's stick to the original logic but loop through ALL matches found in the "tail".
        # We define "tail" as the text after the last "}" (end of a JSON result) or just the end.

        # Actually, let's just process the LAST match. If the user wants parallel,
        # the model should output `<rica>A</rica><rica>B</rica>` and we should process both.

        combined_result = ""

        # We only process the matches that are at the end of the context.
        # i.e. after the last non-whitespace character that is NOT part of a tag.
        # This is hard.

        # Let's just iterate over all matches and check if they look "unprocessed".
        # An unprocessed tag is usually at the end of the context.

        # Let's try to process the last match only, as per current design.
        # To support parallel, we would need to change the loop in `transformers_adapter`
        # to NOT stop immediately?
        # Or we just process all tags found in the `_context`.
        # But we don't want to re-process old tags.

        # Hack: We can check if the tag is followed by a result.
        # But for now, let's just stick to the last one to avoid breaking things,
        # and maybe the user means "parallel threads" not "parallel tools in one thread".
        # User said "同时调用多工具".

        # OK, let's try to find ALL tags that are at the end.
        # We can iterate from the last match backwards until we find something that is not a tag?

        # Let's simplify: Just find the last match. If there are multiple,
        # the previous loop would have caught them?
        # No, `_run_loop` calls this once per generation cycle.

        # If we want parallel, we need to find ALL tags in the current buffer that are not
        # followed by a result.
        # Let's assume any <rica> tag NOT followed by `{` (start of result) or `call_id` is new.

        matches = list(re.finditer(pattern, self._context, re.DOTALL | re.IGNORECASE))
        if not matches:
            return False, None

        # Filter matches that are likely already processed
        candidates = []
        for m in matches:
            end_pos = m.end()
            # Check text after this match
            following_text = self._context[end_pos:].strip()
            # If following text starts with {, it's likely a result.
            if not following_text or (
                not following_text.startswith("{") and not following_text.startswith("[")
            ):
                candidates.append(m)

        if not candidates:
            return False, None

        # Execute all candidates in parallel
        tasks = []
        for match in candidates:
            tag_text = match.group(0)
            tasks.append(self._parse_and_execute(tag_text))

        results = await asyncio.gather(*tasks)

        for res in results:
            if res:
                self._context += res
                await self._emit_token(res)
                combined_result += res

        return True, combined_result

    async def _parse_and_execute(self, tag_text: str) -> str:
        """Helper to parse and execute a single tag."""
        try:
            package_name, route_name, content = parse_rica_tag(tag_text)

            async with self._apps_lock:
                app_instance = self._apps.get(package_name)
                if not app_instance:
                    raise PackageNotFoundError(f"Package '{package_name}' not found")

                application = app_instance.find_route(route_name)
                if not application:
                    raise RouteNotFoundError(f"Route '{route_name}' not found")

            # Special handling for rica/response
            if package_name == "rica" and route_name == "/response":
                await self._emit_response(content)
                return ""  # No result appended for response

            result = await self._execute_tool_call(application, content)

            if isinstance(result, CallBack):
                payload = result.callback
                appended = (
                    json.dumps(payload, ensure_ascii=False)
                    if isinstance(payload, (dict, list))
                    else str(payload)
                )
            else:  # UUID
                appended = json.dumps({"call_id": str(result)}, ensure_ascii=False)

            return appended

        except Exception as e:
            logger.error(f"Tool execution error: {e}", exc_info=True)
            return f"[tool-error]{type(e).__name__}: {e}"
